#!/bin/bash -x
# Copyright (C) 2025 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

wget \
    -O autorestic-1.7.11.tar.gz \
    https://github.com/cupcakearmy/autorestic/archive/v1.7.11/autorestic-1.7.11.tar.gz
